import pandas as pd
import numpy as np
import Bio
from Bio import SeqIO
from Bio.Seq import Seq
from pandarallel import pandarallel
import matplotlib.pyplot as plt
from collections import Counter, defaultdict
from typeguard import typechecked
from datetime import date
from datetime import datetime
import re
from multiprocessing import Pool
from functools import partial
from itertools import repeat
import gzip
import random
from enum import Enum
from typing import Callable
from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict
from typing import Counter as CounterType
from concurrent.futures import ProcessPoolExecutor

from . import crispr_sequence_encoding
from . import crispr_guide_inference
from .crispr_count_processing import get_counterseries_all_results
from ..quality_control.crispr_mapping_quality_control import perform_counts_quality_control
from ..models.mapping_models import GeneralGuideCountType, GeneralMappingInferenceDict
from ..models.mapping_models import WhitelistReporterCountsResult, InferenceResult, CountInput


# TODO: There will probably be some type errors with the DefaultDict when testing on non UMI (since it requires CounterType), so make sure to test with different variations of inputs
@typechecked
def get_whitelist_reporter_counts_with_umi(observed_guide_reporter_umi_counts: GeneralGuideCountType, 
                                           whitelist_guide_reporter_df: pd.DataFrame, 
                                           contains_surrogate:bool = False, 
                                           contains_barcode:bool = False, 
                                           contains_umi:bool = False, 
                                           protospacer_hamming_threshold_strict: Optional[int] = 7, 
                                           surrogate_hamming_threshold_strict: Optional[int] = 2, 
                                           barcode_hamming_threshold_strict: Optional[int] = 2, cores: int=1) -> WhitelistReporterCountsResult:
    
    # Strip all sequences:
    def strip_series(series):
        return series.apply(lambda item : item.rstrip())
    whitelist_guide_reporter_df = whitelist_guide_reporter_df.apply(strip_series, axis=0)

    # Temporary bug fix. Pad sequences so they are all of same length - encoding numpy matrices requires consistent shape. Still pass the original guide table for selecting the matches.     
    def pad_series(series):
        max_surrogate_len = series.apply(len).max()
        return series.apply(lambda item: item.ljust(max_surrogate_len, 'X'))
    padded_whitelist_guide_reporter_df = whitelist_guide_reporter_df.apply(pad_series, axis=0)
    
    #
    # ENCODE THE WHITELISTED SEQUENCES INTO NUMPY MATRICES - THIS IS REQUIRED FOR HAMMING-BASED MAPPING
    #
    encoded_whitelist_protospacer_sequences_series = crispr_sequence_encoding.encode_guide_series(padded_whitelist_guide_reporter_df["protospacer"])
    if contains_surrogate:
        encoded_whitelist_surrogate_sequences_series = crispr_sequence_encoding.encode_guide_series(padded_whitelist_guide_reporter_df["surrogate"])
    if contains_barcode:
        encoded_whitelist_barcode_sequences_series = crispr_sequence_encoding.encode_guide_series(padded_whitelist_guide_reporter_df["barcode"])

    

    
    #
    #    SET THE PROTOSPACER HAMMING THRESHOLD
    #
    # NOTE: Could there be an even more dynamic threshold, that certain mapped guides can have a higher threshold (i.e. those with many editable bases) and other guides can have low hamming (i.e. those with not many editable bases)
    protospacer_hamming_threshold_dynamic = False
    if protospacer_hamming_threshold_strict is None:
        protospacer_hamming_threshold_dynamic = True
        # TODO: Pass in arguments to set this hamming threshold. 
        protospacer_hamming_threshold: int = crispr_guide_inference.determine_hamming_threshold(whitelist_guide_reporter_df["protospacer"], encoded_whitelist_protospacer_sequences_series, sample_count = 100, quantile = 0.05)
    else:
        protospacer_hamming_threshold: int = protospacer_hamming_threshold_strict
    print("Protospacer hamming threshold is " + str(protospacer_hamming_threshold))

    #
    #   SET THE SURROGATE HAMMING THRESHOLD
    #
    if contains_surrogate:
        surrogate_hamming_threshold_dynamic = False
        if surrogate_hamming_threshold_strict is None:
            surrogate_hamming_threshold_dynamic = True
            # TODO: Pass in arguments to set this hamming threshold. 
            surrogate_hamming_threshold: int = crispr_guide_inference.determine_hamming_threshold(whitelist_guide_reporter_df["surrogate"], encoded_whitelist_surrogate_sequences_series, sample_count = 100, quantile = 0.05)
            
        else:
            surrogate_hamming_threshold: int = surrogate_hamming_threshold_strict
        print("Surrogate hamming threshold is " + str(surrogate_hamming_threshold))

    #
    #   SET THE BARCODE HAMMING THRESHOLD
    #
    if contains_barcode:
        barcode_hamming_threshold_dynamic = False
        if barcode_hamming_threshold_strict is  None:
            barcode_hamming_threshold_dynamic = True
            barcode_hamming_threshold: int = crispr_guide_inference.determine_hamming_threshold(whitelist_guide_reporter_df["barcode"], encoded_whitelist_barcode_sequences_series, sample_count = 100, quantile = 0.05)
        else:
            barcode_hamming_threshold: int = barcode_hamming_threshold_strict
        print("Barcode hamming threshold is " + str(barcode_hamming_threshold))



    #
    #   FROM THE OBSERVED SEQUENCES, INFER THE WHITELIST SEQUENCE
    #
    print("Inferring the true guides from observed guides")

    infer_whitelist_sequence_p = partial(crispr_guide_inference.infer_whitelist_sequence,
            whitelist_guide_reporter_df=whitelist_guide_reporter_df,
            contains_surrogate=contains_surrogate,
            contains_barcode=contains_barcode,
            contains_umi=contains_umi,
            encoded_whitelist_protospacer_sequences_series=encoded_whitelist_protospacer_sequences_series,
            encoded_whitelist_surrogate_sequences_series=encoded_whitelist_surrogate_sequences_series,
            encoded_whitelist_barcode_sequences_series=encoded_whitelist_barcode_sequences_series,
            protospacer_hamming_threshold=protospacer_hamming_threshold, 
            surrogate_hamming_threshold=surrogate_hamming_threshold, 
            barcode_hamming_threshold=barcode_hamming_threshold)

    # Perform inference: frpom the observed sequences (previously parsed), infer the true sequence from the whitelist DF.
    observed_guide_reporter_list = observed_guide_reporter_umi_counts.keys()
    inferred_true_reporter_sequences = None
    before_inference_time = datetime.now()
    if cores > 1:
        print(f"Running inference parallelized on {len(observed_guide_reporter_list)} observed seqeunces with cores {cores}")
        with Pool(cores) as pool:
            inferred_true_reporter_sequences = pool.map(
            infer_whitelist_sequence_p,
            observed_guide_reporter_list
           )
    else:
        print(f"Running inference on {len(observed_guide_reporter_list)} observed seqeunces non-parallelized")
        inferred_true_reporter_sequences = [infer_whitelist_sequence_p(observed_guide_reporter) for observed_guide_reporter in observed_guide_reporter_list]
    
    after_inference_time = datetime.now()
    print(f"{(after_inference_time-before_inference_time).seconds} seconds for inference")


    print(f"Mapping inference results of length {len(inferred_true_reporter_sequences)} to the result object")
    # Some organization: Map the inferred result of each observed sequence to a dict with the inferred result and correspoding count
    observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict = defaultdict(dict)
    for observed_guide_reporter_key_index, observed_guide_reporter_key in enumerate(observed_guide_reporter_list):
        observed_guide_reporter_umi_counts_inferred[observed_guide_reporter_key] = InferenceResult(
            observed_value=observed_guide_reporter_umi_counts[observed_guide_reporter_key],
            inferred_value=inferred_true_reporter_sequences[observed_guide_reporter_key_index]
        )
    
    after_inference_processing_time = datetime.now()
    print(f"{(after_inference_processing_time-after_inference_time).seconds} seconds for inference processing")
    print("Completed inference")

    # GET THE MAPPED COUNT SERIES BASED ON THE INFERENCE RESULTS
    print("Prepare the processed count series ")
    # Count
    all_match_set_whitelist_reporter_counter_series_results = get_counterseries_all_results(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_barcode, contains_surrogate, contains_umi)

    after_counterseries_time = datetime.now()
    print(f"{(after_counterseries_time-after_inference_processing_time).seconds} seconds for counter series generation")

    print("Preparing quality control")
    quality_control_result = perform_counts_quality_control(observed_guide_reporter_umi_counts_inferred, contains_umi, contains_surrogate, contains_barcode)
    
    after_qualitycontrol_time = datetime.now()
    print(f"{(after_qualitycontrol_time-after_counterseries_time).seconds} seconds for quality control")

    count_input = CountInput(whitelist_guide_reporter_df=whitelist_guide_reporter_df,
            contains_surrogate=contains_surrogate,
            contains_barcode=contains_barcode,
            contains_umi=contains_umi,
            protospacer_hamming_threshold_strict=protospacer_hamming_threshold_strict,
            surrogate_hamming_threshold_strict=surrogate_hamming_threshold_strict,
            barcode_hamming_threshold_strict=barcode_hamming_threshold_strict)
    
    return WhitelistReporterCountsResult(all_match_set_whitelist_reporter_counter_series_results=all_match_set_whitelist_reporter_counter_series_results, observed_guide_reporter_umi_counts_inferred=observed_guide_reporter_umi_counts_inferred, quality_control_result=quality_control_result, count_input=count_input)


import pandas as pd
import anndata as ad
from typing import Optional, List
import numpy as np

def initialize_anndata(sample_info_df: pd.DataFrame, 
                       guide_library_df: pd.DataFrame,
                       count_df: Optional[pd.DataFrame] = None, count_series_list: Optional[List[pd.Series]] = None) -> ad.AnnData:
    # Prepare sample_info_df
    sample_info_df = sample_info_df.reset_index(drop=True)

    # Prepare count_df
    if count_df is None:
        if count_series_list is not None:
            count_df = pd.DataFrame(count_series_list).transpose()
            count_df.columns = count_df.index
        else:
            raise Exception("Must provide processed count_df or unprocessed count_series_list")
    
    # Set guide_library_df index
    guide_columns = ['protospacer', 'surrogate', 'barcode']
    available_columns = [col for col in guide_columns if col in guide_library_df.columns]
    guide_library_df.index = guide_library_df["protospacer"] + "_" +  guide_library_df["surrogate"] + "_" + guide_library_df["barcode"]
    guide_library_df.index = guide_library_df[available_columns].apply(lambda row: '_'.join(row.values.astype(str)), axis=1)

    # Set count_df indices
    count_df_index = count_df.index.to_frame()
    count_df.index = count_df_index[available_columns].apply(lambda row: '_'.join(row.values.astype(str)), axis=1)
    count_df = count_df.loc[guide_library_df.index] # Filter count_df from available guides
    count_df.columns = sample_info_df.index

    # Initialize AnnData
    count_anndata = ad.AnnData(var=guide_library_df, obs=sample_info_df, X=count_df.transpose())

    return count_anndata

def add_lfc_layer(anndata: ad.AnnData, 
                  groupby_columns_list: List[str], 
                  enriched_population_label: str, 
                  baseline_population_label: str, 
                  population_column_name: str) -> ad.AnnData:
    lfc_dict = {} 
    for group in anndata.obs.groupby(groupby_columns_list):
        enriched_pop_sample_df = group[1][group[1][population_column_name] == enriched_population_label]
        assert enriched_pop_sample_df.shape[0] == 1, f"Multiple samples matching enriched condition in group {group[1]}"
        enriched_pop_sample = enriched_pop_sample_df.iloc[0,:]
        
        baseline_pop_sample_df = group[1][group[1][population_column_name] == baseline_population_label]
        assert baseline_pop_sample_df.shape[0] == 1, f"Multiple samples matching enriched condition in group {group[1]}"
        baseline_pop_sample = baseline_pop_sample_df.iloc[0,:]

        enriched_anndata = anndata[int(enriched_pop_sample.name), :] # DEVELOPER NOTE: Will thow error of indices of samples are changed from ints
        baseline_anndata = anndata[int(baseline_pop_sample.name), :] # DEVELOPER NOTE: Will thow error of indices of samples are changed from ints

        enriched_X = enriched_anndata.X[0]
        baseline_X = baseline_anndata.X[0]

        enriched_X_safetargeting_total = enriched_anndata[:, enriched_anndata.var["type"] == "safe-targeting"].X[0].sum()
        baseline_X_safetargeting_total = baseline_anndata[:, baseline_anndata.var["type"] == "safe-targeting"].X[0].sum()

        baseline_normalization_constant = enriched_X_safetargeting_total / baseline_X_safetargeting_total

        baseline_X_normalized = baseline_X * baseline_normalization_constant

        enriched_lfc = np.log(np.divide(enriched_X + 1, baseline_X_normalized + 1))
        baseline_lfc = np.log(np.divide(baseline_X_normalized + 1, enriched_X + 1))

        lfc_dict.update({int(enriched_pop_sample.name): enriched_lfc})
        lfc_dict.update({int(baseline_pop_sample.name): baseline_lfc})
    
    # NOTE: LFC layer for all samples must be added at once
    anndata.layers["lfc"] = np.asarray([lfc_dict[idx] for idx in range(0, anndata.shape[0])])
    return anndata


from dataclasses import dataclass
from typing import Counter as CounterType
from typing import Tuple, Optional, DefaultDict, Dict
from .mapping_models import GeneralAlleleCountSeriesDict
import pandas as pd

@dataclass
class MatchSetWhitelistReporterObservedSequenceCounterSeriesResults:
    
    # Storing as a dictionary 
    ambiguous_ignored_umi_noncollapsed_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None
    ambiguous_ignored_umi_collapsed_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None
    ambiguous_ignored_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None

    ambiguous_accepted_umi_noncollapsed_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None
    ambiguous_accepted_umi_collapsed_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None
    ambiguous_accepted_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None

    ambiguous_spread_umi_noncollapsed_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None
    ambiguous_spread_umi_collapsed_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None
    ambiguous_spread_alleleseries_dict : Optional[GeneralAlleleCountSeriesDict] = None
        
    # Storing as a dataframe
    ambiguous_ignored_umi_noncollapsed_allele_df : Optional[pd.DataFrame] = None
    ambiguous_ignored_umi_collapsed_allele_df : Optional[pd.DataFrame] = None
    ambiguous_ignored_allele_df : Optional[pd.DataFrame] = None

    ambiguous_accepted_umi_noncollapsed_allele_df : Optional[pd.DataFrame] = None
    ambiguous_accepted_umi_collapsed_allele_df : Optional[pd.DataFrame] = None
    ambiguous_accepted_allele_df : Optional[pd.DataFrame] = None

    ambiguous_spread_umi_noncollapsed_allele_df : Optional[pd.DataFrame] = None
    ambiguous_spread_umi_collapsed_allele_df : Optional[pd.DataFrame] = None
    ambiguous_spread_allele_df : Optional[pd.DataFrame] = None

    # This ensures that any empty series are kept at None
    def __setattr__(self, name, value):
        if isinstance(value, dict):
            if len(value) == 0: # If no items in dictionary, just return None
                super().__setattr__(name, None)
            else:
                super().__setattr__(name, value)
        elif isinstance(value, pd.DataFrame):
            if value.empty: # If no items in dataframe, just return None
                super().__setattr__(name, None)
            else:
                super().__setattr__(name, value)
        else:
            # Set the attribute
            super().__setattr__(name, value)

@dataclass
class ObservedSequenceMutationProfile:
    linked_mutations_whitelist_reporter_dict: Dict[Tuple[str, Optional[str], Optional[str]], pd.DataFrame] = None
    all_observed_protospacer_unlinked_mutations_df: pd.DataFrame = None
    all_observed_surrogate_unlinked_mutations_df: Optional[pd.DataFrame] = None
    all_observed_barcode_unlinked_mutations_df: Optional[pd.DataFrame] = None
    
@dataclass
class MatchSetWhitelistReporterObservedSequenceMutationProfiles:
    ambiguous_ignored_umi_noncollapsed_mutations : Optional[ObservedSequenceMutationProfile] = None
    ambiguous_ignored_umi_collapsed_mutations : Optional[ObservedSequenceMutationProfile] = None
    ambiguous_ignored_mutations : Optional[ObservedSequenceMutationProfile] = None

    ambiguous_accepted_umi_noncollapsed_mutations : Optional[ObservedSequenceMutationProfile] = None
    ambiguous_accepted_umi_collapsed_mutations : Optional[ObservedSequenceMutationProfile] = None
    ambiguous_accepted_mutations : Optional[ObservedSequenceMutationProfile] = None

    ambiguous_spread_umi_noncollapsed_mutations : Optional[ObservedSequenceMutationProfile] = None
    ambiguous_spread_umi_collapsed_mutations : Optional[ObservedSequenceMutationProfile] = None
    ambiguous_spread_mutations : Optional[ObservedSequenceMutationProfile] = None

@dataclass
class LinkedMutationCounters:
    protospacer_total_mutation_counter: CounterType
    surrogate_total_mutation_counter: Optional[CounterType]
    barcode_total_mutation_counter: Optional[CounterType]
    


from dataclasses import dataclass
from typing import Optional, DefaultDict
from .error_models import GuideCountError

@dataclass
class SingleInferenceQualityControlResult:
    non_error_dict: Optional[DefaultDict] = None
    num_non_error_umi_noncollapsed_counts: Optional[int] = None
    num_non_error_umi_collapsed_counts: Optional[int] = None
    num_non_error_counts: Optional[int] = None
    num_total_umi_noncollapsed_counts: Optional[int] = None
    num_total_umi_collapsed_counts: Optional[int] = None
    num_total_counts: Optional[int] = None
    guide_count_error_type_umi_noncollapsed_count: DefaultDict[GuideCountError, int] = None
    guide_count_error_type_umi_collapsed_count: DefaultDict[GuideCountError, int] = None
    guide_count_error_type_count: DefaultDict[GuideCountError, int] = None

@dataclass
class MatchSetSingleInferenceQualityControlResult(SingleInferenceQualityControlResult):
    # TODO: In the future, can add any QC that is specific to the singe match set result type
    pass

@dataclass
class SurrogateProtospacerMismatchSingleInferenceQualityControlResult(SingleInferenceQualityControlResult):
    # TODO: In the future, can add any QC that is specific to the surrogate_protospacer mismatch result type
    pass

@dataclass
class QualityControlResult:
    protospacer_match: Optional[MatchSetSingleInferenceQualityControlResult] = None
    protospacer_match_surrogate_match_barcode_match: Optional[MatchSetSingleInferenceQualityControlResult] = None
    protospacer_match_surrogate_match: Optional[MatchSetSingleInferenceQualityControlResult] = None
    protospacer_match_barcode_match: Optional[MatchSetSingleInferenceQualityControlResult] = None
    protospacer_mismatch_surrogate_match_barcode_match: Optional[SurrogateProtospacerMismatchSingleInferenceQualityControlResult] = None
    protospacer_mismatch_surrogate_match: Optional[SurrogateProtospacerMismatchSingleInferenceQualityControlResult] = None